import pytz
from peewee import CharField, TimestampField, ForeignKeyField, IntegerField, fn
from ..dbmodels.core import BaseModel
from datetime import datetime
from .users import Users
from ..modules.users.users import User

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"


class Events(BaseModel):
    r"""
    Database model for System Events.

    Events track user actions and system changes.
    """
    
    timestamp = TimestampField(utc=True)
    message = CharField(max_length=256)
    description = CharField(max_length=256, null=True)
    classification = CharField(max_length=128, null=True)
    priority = IntegerField(null=True)
    criticity = IntegerField(null=True)
    user = ForeignKeyField(Users, backref='events')

    @classmethod
    def create(
        cls, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        priority:int=None,
        criticity:int=None,
        timestamp:datetime=None
        )->tuple:
        r"""
        Creates a new event record.

        **Parameters:**

        * **message** (str): Event message.
        * **user** (User): User associated with the event.
        * **description** (str, optional): Detailed description.
        * **classification** (str, optional): Event category.
        * **priority** (int, optional): Priority level.
        * **criticity** (int, optional): Criticity level.
        * **timestamp** (datetime, optional): Time of event.

        **Returns:**

        * **tuple**: (Query object, status message)
        """

        if not isinstance(user, User):

            return None, f"User {user} - {type(user)} must be an User Object"
        
        user = Users.read_by_username(username=user.username) 

        if not timestamp:

            timestamp = datetime.now(pytz.UTC)
        
        if not isinstance(timestamp, datetime):

            return None, f"Timestamp must be a datetime Object"
        
        # Ensure timestamp is timezone-aware and in UTC
        if timestamp.tzinfo is None:
            # If naive, assume it's UTC
            timestamp = pytz.UTC.localize(timestamp)
        else:
            # If timezone-aware, convert to UTC
            timestamp = timestamp.astimezone(pytz.UTC)

        query = cls(
            message=message,
            user=user,
            description=description,
            classification=classification,
            priority=priority,
            criticity=criticity,
            timestamp=timestamp
        )
        query.save()

        return query, f"Event creation successful"
    
    @classmethod
    def read_lasts(cls, lasts:int=1):
        r"""
        Retrieves the last N events.

        **Parameters:**

        * **lasts** (int): Number of events to retrieve.

        **Returns:**

        * **list**: List of serialized event dictionaries.
        """
        events = cls.select().order_by(cls.id.desc()).limit(lasts)

        return [event.serialize() for event in events]
    
    @classmethod
    def filter_by(
        cls, 
        usernames:list[str]=None,
        priorities:list[int]=None,
        criticities:list[int]=None,
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        description:str="",
        message:str="",
        classification:str="",
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):
        r"""
        Filters events based on criteria with pagination.

        **Parameters:**

        * **usernames** (list[str]): Filter by user.
        * **priorities** (list[int]): Filter by priority.
        * **criticities** (list[int]): Filter by criticity.
        * **message**, **description**, **classification**: Text search.
        * **greater_than_timestamp**, **less_than_timestamp**: Time range.
        * **page**, **limit**: Pagination.

        **Returns:**

        * **dict**: {data: list, pagination: dict}
        """
        import math
        _timezone = pytz.timezone(timezone)
        query = cls.select()
        
        if usernames:
            subquery = Users.select(Users.id).where(Users.username.in_(usernames))
            query = query.join(Users).where(Users.id.in_(subquery))
        
        if priorities:
            query = query.where(cls.priority.in_(priorities))
            
        if criticities:
            query = query.where(cls.criticity.in_(criticities))
            
        if description:
            query = query.where(fn.LOWER(cls.description).contains(description.lower()))
            
        if message:
            query = query.where(fn.LOWER(cls.message).contains(message.lower()))
            
        if classification:
            query = query.where(fn.LOWER(cls.classification).contains(classification.lower()))
            
        if greater_than_timestamp:
            # If it's already a datetime object (naive UTC from endpoint), use it directly
            if isinstance(greater_than_timestamp, datetime):
                if greater_than_timestamp.tzinfo is not None:
                    # Convert to naive UTC
                    dt_utc = greater_than_timestamp.astimezone(pytz.UTC)
                    greater_than_timestamp = dt_utc.replace(tzinfo=None)
            else:
                # Legacy: parse string and convert from user timezone to UTC
                try:
                    if '.' in str(greater_than_timestamp):
                        parts = str(greater_than_timestamp).split('.')
                        base_time = parts[0]
                        microseconds = parts[1] if len(parts) > 1 else '0'
                        microseconds = microseconds.ljust(6, '0')[:6]
                        formatted_str = f"{base_time}.{microseconds}"
                        dt_naive = datetime.strptime(formatted_str, '%Y-%m-%d %H:%M:%S.%f')
                    else:
                        dt_naive = datetime.strptime(str(greater_than_timestamp), '%Y-%m-%d %H:%M:%S')
                    dt_local = _timezone.localize(dt_naive)
                    dt_utc = dt_local.astimezone(pytz.UTC)
                    greater_than_timestamp = dt_utc.replace(tzinfo=None)
                except ValueError:
                    # Try ISO format as fallback
                    dt = datetime.fromisoformat(str(greater_than_timestamp).replace('Z', '+00:00'))
                    if dt.tzinfo is not None:
                        dt = dt.astimezone(pytz.UTC)
                    greater_than_timestamp = dt.replace(tzinfo=None)
            
            query = query.where(cls.timestamp > greater_than_timestamp)
            
        if less_than_timestamp:
            # If it's already a datetime object (naive UTC from endpoint), use it directly
            if isinstance(less_than_timestamp, datetime):
                if less_than_timestamp.tzinfo is not None:
                    # Convert to naive UTC
                    dt_utc = less_than_timestamp.astimezone(pytz.UTC)
                    less_than_timestamp = dt_utc.replace(tzinfo=None)
            else:
                # Legacy: parse string and convert from user timezone to UTC
                try:
                    if '.' in str(less_than_timestamp):
                        parts = str(less_than_timestamp).split('.')
                        base_time = parts[0]
                        microseconds = parts[1] if len(parts) > 1 else '0'
                        microseconds = microseconds.ljust(6, '0')[:6]
                        formatted_str = f"{base_time}.{microseconds}"
                        dt_naive = datetime.strptime(formatted_str, '%Y-%m-%d %H:%M:%S.%f')
                    else:
                        dt_naive = datetime.strptime(str(less_than_timestamp), '%Y-%m-%d %H:%M:%S')
                    dt_local = _timezone.localize(dt_naive)
                    dt_utc = dt_local.astimezone(pytz.UTC)
                    less_than_timestamp = dt_utc.replace(tzinfo=None)
                except ValueError:
                    # Try ISO format as fallback
                    dt = datetime.fromisoformat(str(less_than_timestamp).replace('Z', '+00:00'))
                    if dt.tzinfo is not None:
                        dt = dt.astimezone(pytz.UTC)
                    less_than_timestamp = dt.replace(tzinfo=None)
            
            query = query.where(cls.timestamp < less_than_timestamp)
            
        query = query.order_by(cls.id.desc())

        total_records = query.count()
        
        if limit <= 0: limit = 20
        if page <= 0: page = 1
        
        total_pages = math.ceil(total_records / limit)
        if total_pages == 0: total_pages = 1
        
        has_next = page < total_pages
        has_prev = page > 1
        
        paginated_query = query.paginate(page, limit)
        
        # Serialize events with the specified timezone
        data = [event.serialize(timezone=timezone) for event in paginated_query]
        
        return {
            "data": data,
            "pagination": {
                "page": page,
                "limit": limit,
                "total_records": total_records,
                "total_pages": total_pages,
                "has_next": has_next,
                "has_prev": has_prev
            }
        }
    
    @classmethod
    def get_comments(cls, id:int):
        r"""
        Retrieves comments (logs) associated with an event.
        """
        query = cls.read(id=id)

        return [comment.serialize() for comment in query.logs]

    def serialize(self, timezone=None)-> dict:
        r"""
        Serializes the event record.
        
        **Parameters:**
        
        * **timezone** (str, optional): Timezone to convert timestamp to. If None, uses default TIMEZONE.
        """
        from .. import TIMEZONE, MANUFACTURER, SEGMENT
        timestamp = self.timestamp
        if timestamp:
            # Convert to specified timezone or default TIMEZONE
            target_tz = pytz.timezone(timezone) if timezone else TIMEZONE
            # If timestamp is naive, assume it's UTC
            if timestamp.tzinfo is None:
                timestamp = pytz.UTC.localize(timestamp)
            timestamp = timestamp.astimezone(target_tz)
            timestamp = timestamp.strftime(DATETIME_FORMAT)

        return {
            "id": self.id,
            "timestamp": timestamp,
            "user": self.user.serialize(),
            "message": self.message,
            "description": self.description,
            "classification": self.classification,
            "priority": self.priority,
            "criticity": self.criticity,
            "segment": SEGMENT,
            "manufacturer": MANUFACTURER,
            "has_comments": True if self.logs else False
        }



from peewee import ForeignKeyField, FloatField
from .core import BaseModel
from .tags import Segment


class LinearReferencingGeospatial(BaseModel):
    r"""
    Database model for geospatial linear-referencing points of a pipeline segment.
    """

    segment = ForeignKeyField(Segment, backref='linear_referencing_points', on_delete='CASCADE')
    kp = FloatField()
    latitude = FloatField()
    longitude = FloatField()
    elevation = FloatField(null=True)

    class Meta:
        indexes = (
            (('segment', 'kp'), True),
        )

    @classmethod
    def create(
        cls,
        segment_name:str,
        kp:float,
        latitude:float,
        longitude:float,
        elevation:float=None
    ):
        r"""
        Creates a new linear-referencing geospatial point.
        """
        segment_obj = Segment.read_by_name(name=segment_name)
        if segment_obj is None:
            return None, f"Segment {segment_name} does not exist into database"

        duplicated = cls.get_or_none((cls.segment == segment_obj) & (cls.kp == kp))
        if duplicated is not None:
            return None, f"Duplicated linear referencing point: {segment_name} KP {kp}"

        query = cls(
            segment=segment_obj,
            kp=kp,
            latitude=latitude,
            longitude=longitude,
            elevation=elevation
        )
        query.save()
        return query, "Linear referencing geospatial point created successfully"

    @classmethod
    def read_by_segment_name(cls, segment_name:str):
        r"""
        Retrieves all points by segment name ordered by KP ascending.
        """
        segment_obj = Segment.read_by_name(name=segment_name)
        if segment_obj is None:
            return []
        query = cls.select().where(cls.segment == segment_obj).order_by(cls.kp.asc())
        return query

    @classmethod
    def delete_by_id(cls, id:int):
        r"""
        Deletes a point by its numeric ID.
        """
        if not cls.id_exists(id):
            return 0
        query = cls.delete().where(cls.id == id)
        return query.execute()

    @classmethod
    def interpolate_by_segment_and_kp(cls, segment_name:str, kp:float):
        r"""
        Retrieves geospatial data by segment and KP using linear interpolation if needed.
        """
        segment_obj = Segment.read_by_name(name=segment_name)
        if segment_obj is None:
            return None, f"Segment {segment_name} does not exist into database"

        exact = cls.get_or_none((cls.segment == segment_obj) & (cls.kp == kp))
        if exact is not None:
            data = exact.serialize()
            data["interpolated"] = False
            return data, "Exact KP match"

        lower = (
            cls.select()
            .where((cls.segment == segment_obj) & (cls.kp < kp))
            .order_by(cls.kp.desc())
            .first()
        )
        upper = (
            cls.select()
            .where((cls.segment == segment_obj) & (cls.kp > kp))
            .order_by(cls.kp.asc())
            .first()
        )

        if lower is None and upper is None:
            return None, f"No geospatial points configured for segment {segment_name}"

        if lower is None:
            data = upper.serialize()
            data["requested_kp"] = kp
            data["interpolated"] = False
            data["interpolation_note"] = "Requested KP below available range; nearest upper point returned"
            return data, "Nearest point returned"

        if upper is None:
            data = lower.serialize()
            data["requested_kp"] = kp
            data["interpolated"] = False
            data["interpolation_note"] = "Requested KP above available range; nearest lower point returned"
            return data, "Nearest point returned"

        kp_span = upper.kp - lower.kp
        if kp_span == 0:
            return None, "Invalid interpolation range (duplicated KP points)"

        ratio = (kp - lower.kp) / kp_span
        latitude = lower.latitude + (upper.latitude - lower.latitude) * ratio
        longitude = lower.longitude + (upper.longitude - lower.longitude) * ratio
        elevation = None
        if lower.elevation is not None and upper.elevation is not None:
            elevation = lower.elevation + (upper.elevation - lower.elevation) * ratio

        return {
            "id": None,
            "segment_id": segment_obj.id,
            "segment_name": segment_obj.name,
            "kp": kp,
            "latitude": latitude,
            "longitude": longitude,
            "elevation": elevation,
            "interpolated": True,
            "interpolation_bounds": {
                "lower_kp": lower.kp,
                "upper_kp": upper.kp
            }
        }, "Interpolated KP location"

    def serialize(self):
        r"""
        Serializes the linear-referencing geospatial point.
        """
        return {
            "id": self.id,
            "segment_id": self.segment.id,
            "segment_name": self.segment.name,
            "kp": self.kp,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "elevation": self.elevation
        }



from ....extensions.api import api


def init_app():
    from .linear_referencing_geospatial import ns as ns_linear_referencing

    api.add_namespace(ns_linear_referencing, path="/linear-referencing-geospatial")



import pytz
from peewee import CharField, TimestampField, ForeignKeyField, fn
from ..dbmodels.core import BaseModel
from datetime import datetime
from .users import Users
from .events import Events
from .alarms import AlarmSummary, Alarms
from ..modules.users.users import User

DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S.%f"


class Logs(BaseModel):
    r"""
    Database model for Application Logs.

    Logs store system messages, errors, and informational records, optionally linked to
    events or alarms.
    """
    
    timestamp = TimestampField(utc=True)
    message = CharField(max_length=256)
    description = CharField(max_length=256, null=True)
    classification = CharField(max_length=128, null=True)
    user = ForeignKeyField(Users, backref='logs', on_delete='CASCADE')
    alarm = ForeignKeyField(AlarmSummary, null=True, backref='logs', on_delete='CASCADE')
    event = ForeignKeyField(Events, null=True, backref='logs', on_delete='CASCADE')

    @classmethod
    def create(
        cls, 
        message:str, 
        user:User, 
        description:str=None, 
        classification:str=None,
        alarm_summary_id:int=None,
        event_id:int=None,
        timestamp:datetime=None
        )->tuple:
        r"""
        Creates a new log entry.

        **Parameters:**

        * **message** (str): Log content.
        * **user** (User): User creating the log.
        * **description** (str, optional): Additional details.
        * **classification** (str, optional): Log type/category.
        * **alarm_summary_id** (int, optional): Link to an alarm summary entry.
        * **event_id** (int, optional): Link to an event.
        * **timestamp** (datetime, optional): Log time.

        **Returns:**

        * **tuple**: (Query object, status message)
        """
        if not isinstance(user, User):

            return None, f"User {user} - {type(user)} must be an User Object"
        
        _user = Users.read_by_username(username=user.username) 

        if not timestamp:

            timestamp = datetime.now(pytz.UTC)
        
        if not isinstance(timestamp, datetime):

            return None, f"Timestamp must be a datetime Object"
        
        # Ensure timestamp is timezone-aware and in UTC
        if timestamp.tzinfo is None:
            # If naive, assume it's UTC
            timestamp = pytz.UTC.localize(timestamp)
        else:
            # If timezone-aware, convert to UTC
            timestamp = timestamp.astimezone(pytz.UTC)
        
        query = cls(
            message=message,
            user=_user,
            description=description,
            classification=classification,
            timestamp=timestamp,
            event=Events.get_or_none(id=event_id),
            alarm=AlarmSummary.get_or_none(id=alarm_summary_id)
        )
        query.save()

        return query, f"Event creation successful"
    
    @classmethod
    def read_lasts(cls, lasts:int=1):
        r"""
        Retrieves the last N logs.

        **Parameters:**

        * **lasts** (int): Number of logs to retrieve.

        **Returns:**

        * **list**: List of serialized log dictionaries.
        """
        logs = cls.select().order_by(cls.id.desc()).limit(lasts)

        return [log.serialize() for log in logs]
    
    @classmethod
    def filter_by(
        cls, 
        usernames:list[str]=None,
        alarm_names:list[str]=None,
        event_ids:list[int]=None,
        description:str="",
        message:str="",
        classification:str="",
        greater_than_timestamp:datetime=None,
        less_than_timestamp:datetime=None,
        timezone:str='UTC',
        page:int=1,
        limit:int=20
        ):
        r"""
        Filters logs based on various criteria with pagination.

        **Parameters:**

        * **usernames** (list[str]): Filter by user.
        * **alarm_names** (list[str]): Filter by linked alarm name.
        * **event_ids** (list[int]): Filter by linked event ID.
        * **message**, **description**, **classification**: Text search.
        * **greater_than_timestamp**, **less_than_timestamp**: Time range.
        * **page**, **limit**: Pagination control.

        **Returns:**

        * **dict**: {data: list, pagination: dict}
        """
        import math
        _timezone = pytz.timezone(timezone)
        query = cls.select()
        
        if usernames:
            subquery = Users.select(Users.id).where(Users.username.in_(usernames))
            query = query.join(Users).where(Users.id.in_(subquery))
        
        if event_ids:
            subquery = Events.select(Events.id).where(Events.id.in_(event_ids))
            query = query.join(Events).where(Events.id.in_(subquery))
        
        if alarm_names:
            subquery = Alarms.select(Alarms.id).where(Alarms.name.in_(alarm_names))
            alarm_subquery = AlarmSummary.select(AlarmSummary.id).join(Alarms).where(Alarms.id.in_(subquery))
            query = query.join(AlarmSummary).where(AlarmSummary.id.in_(alarm_subquery))
        
        if description:
            query = query.where(fn.LOWER(cls.description).contains(description.lower()))
        
        if message:
            query = query.where(fn.LOWER(cls.message).contains(message.lower()))
        
        if classification:
            query = query.where(fn.LOWER(cls.classification).contains(classification.lower()))
        
        if greater_than_timestamp:
            # If it's already a datetime object (naive UTC from endpoint), use it directly
            if isinstance(greater_than_timestamp, datetime):
                if greater_than_timestamp.tzinfo is not None:
                    # Convert to naive UTC
                    dt_utc = greater_than_timestamp.astimezone(pytz.UTC)
                    greater_than_timestamp = dt_utc.replace(tzinfo=None)
            else:
                # Legacy: parse string and convert from user timezone to UTC
                try:
                    if '.' in str(greater_than_timestamp):
                        parts = str(greater_than_timestamp).split('.')
                        base_time = parts[0]
                        microseconds = parts[1] if len(parts) > 1 else '0'
                        microseconds = microseconds.ljust(6, '0')[:6]
                        formatted_str = f"{base_time}.{microseconds}"
                        dt_naive = datetime.strptime(formatted_str, '%Y-%m-%d %H:%M:%S.%f')
                    else:
                        dt_naive = datetime.strptime(str(greater_than_timestamp), '%Y-%m-%d %H:%M:%S')
                    dt_local = _timezone.localize(dt_naive)
                    dt_utc = dt_local.astimezone(pytz.UTC)
                    greater_than_timestamp = dt_utc.replace(tzinfo=None)
                except ValueError:
                    # Try ISO format as fallback
                    dt = datetime.fromisoformat(str(greater_than_timestamp).replace('Z', '+00:00'))
                    if dt.tzinfo is not None:
                        dt = dt.astimezone(pytz.UTC)
                    greater_than_timestamp = dt.replace(tzinfo=None)
            
            query = query.where(cls.timestamp > greater_than_timestamp)
        
        if less_than_timestamp:
            # If it's already a datetime object (naive UTC from endpoint), use it directly
            if isinstance(less_than_timestamp, datetime):
                if less_than_timestamp.tzinfo is not None:
                    # Convert to naive UTC
                    dt_utc = less_than_timestamp.astimezone(pytz.UTC)
                    less_than_timestamp = dt_utc.replace(tzinfo=None)
            else:
                # Legacy: parse string and convert from user timezone to UTC
                try:
                    if '.' in str(less_than_timestamp):
                        parts = str(less_than_timestamp).split('.')
                        base_time = parts[0]
                        microseconds = parts[1] if len(parts) > 1 else '0'
                        microseconds = microseconds.ljust(6, '0')[:6]
                        formatted_str = f"{base_time}.{microseconds}"
                        dt_naive = datetime.strptime(formatted_str, '%Y-%m-%d %H:%M:%S.%f')
                    else:
                        dt_naive = datetime.strptime(str(less_than_timestamp), '%Y-%m-%d %H:%M:%S')
                    dt_local = _timezone.localize(dt_naive)
                    dt_utc = dt_local.astimezone(pytz.UTC)
                    less_than_timestamp = dt_utc.replace(tzinfo=None)
                except ValueError:
                    # Try ISO format as fallback
                    dt = datetime.fromisoformat(str(less_than_timestamp).replace('Z', '+00:00'))
                    if dt.tzinfo is not None:
                        dt = dt.astimezone(pytz.UTC)
                    less_than_timestamp = dt.replace(tzinfo=None)
            
            query = query.where(cls.timestamp < less_than_timestamp)
        
        query = query.order_by(cls.id.desc())

        total_records = query.count()
        
        if limit <= 0: limit = 20
        if page <= 0: page = 1
        
        total_pages = math.ceil(total_records / limit)
        if total_pages == 0: total_pages = 1
        
        has_next = page < total_pages
        has_prev = page > 1
        
        paginated_query = query.paginate(page, limit)
        
        # Serialize logs with the specified timezone
        data = [log.serialize(timezone=timezone) for log in paginated_query]
        
        return {
            "data": data,
            "pagination": {
                "page": page,
                "limit": limit,
                "total_records": total_records,
                "total_pages": total_pages,
                "has_next": has_next,
                "has_prev": has_prev
            }
        }

    def serialize(self, timezone=None)-> dict:
        r"""
        Serializes the log record.
        
        **Parameters:**
        
        * **timezone** (str, optional): Timezone to convert timestamp to. If None, uses default TIMEZONE.
        """
        from .. import MANUFACTURER, SEGMENT, TIMEZONE
        timestamp = self.timestamp
        if timestamp:
            # Convert to specified timezone or default TIMEZONE
            target_tz = pytz.timezone(timezone) if timezone else TIMEZONE
            # If timestamp is naive, assume it's UTC
            if timestamp.tzinfo is None:
                timestamp = pytz.UTC.localize(timestamp)
            timestamp = timestamp.astimezone(target_tz)
            timestamp = timestamp.strftime(DATETIME_FORMAT)

        _event = None
        if self.event:

            _event = self.event.serialize()

        _alarm = None
        if self.alarm:

            _alarm = self.alarm.serialize()

        return {
            "id": self.id,
            "timestamp": timestamp,
            "user": self.user.serialize(),
            "message": self.message,
            "description": self.description,
            "classification": self.classification,
            "event": _event,
            "alarm": _alarm,
            "segment": SEGMENT,
            "manufacturer": MANUFACTURER
        }


from ....extensions.api import api


def init_app():

    from .clients import ns as ns_clients
    from .server import ns as ns_server

    api.add_namespace(ns_clients, path="/opcua/clients")
    api.add_namespace(ns_server, path="/opcua/server")




from flask_restx import Namespace, Resource, fields
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api

ns = Namespace('Database', description='Database Configuration and Connection Management')
app = PyAutomation()

# Modelo para la configuración de conexión a la base de datos
db_connect_model = api.model("db_connect_model", {
    'dbtype': fields.String(required=True, description='Database type: sqlite, mysql, or postgresql'),
    'dbfile': fields.String(required=False, description='Database filename (for SQLite only)'),
    'user': fields.String(required=False, description='Database user (for MySQL/PostgreSQL)'),
    'password': fields.String(required=False, description='Database password (for MySQL/PostgreSQL)'),
    'host': fields.String(required=False, description='Database host (for MySQL/PostgreSQL)'),
    'port': fields.Integer(required=False, description='Database port (for MySQL/PostgreSQL)'),
    'name': fields.String(required=False, description='Database name (for MySQL/PostgreSQL)'),
    'reload': fields.Boolean(required=False, default=False, description='Reload tags into machine after connection')
})

@ns.route('/config')
class DatabaseConfigResource(Resource):

    # @api.doc(security='apikey', description="Retrieves the current database configuration.")
    @api.doc(security=None, description="Retrieves the current database configuration.")
    @api.response(200, "Success")
    @api.response(404, "Configuration not found")
    # @Api.token_required(auth=True)
    def get(self):
        r"""
        Get database configuration.

        Retrieves the current database configuration from db_config.json.
        """
        config = app.get_db_config()
        if config:
            return config, 200
        return {"message": "Database configuration not found"}, 404

@ns.route('/connected')
class DatabaseConnectedResource(Resource):

    @api.doc(security='apikey', description="Checks if the database is currently connected.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self):
        r"""
        Check database connection status.

        Returns whether the database is currently connected.
        """
        is_connected = app.is_db_connected()
        return {"connected": is_connected}, 200

@ns.route('/connect')
class DatabaseConnectResource(Resource):

    # @api.doc(security='apikey', description="Connects to the database with the provided configuration.")
    @api.doc(security=None, description="Connects to the database with the provided configuration.")
    @api.response(200, "Connection successful")
    @api.response(400, "Connection failed")
    # @Api.token_required(auth=True)
    @ns.expect(db_connect_model)
    def post(self):
        r"""
        Connect to database.

        Sets the database configuration and establishes a connection.
        For SQLite, only 'dbtype' and 'dbfile' are required.
        For MySQL/PostgreSQL, 'dbtype', 'user', 'password', 'host', 'port', and 'name' are required.
        """
        try:
            # Extraer parámetros del payload
            dbtype = api.payload.get('dbtype', 'sqlite')
            dbfile = api.payload.get('dbfile', 'app.db')
            user = api.payload.get('user')
            password = api.payload.get('password')
            host = api.payload.get('host', '127.0.0.1')
            port = api.payload.get('port', 5432)
            name = api.payload.get('name', 'app_db')
            reload = api.payload.get('reload', False)

            # Validar parámetros según el tipo de base de datos
            if dbtype.lower() == 'sqlite':
                # Para SQLite solo necesitamos dbtype y dbfile
                app.set_db_config(
                    dbtype=dbtype,
                    dbfile=dbfile
                )
            elif dbtype.lower() in ['mysql', 'postgresql']:
                # Para MySQL/PostgreSQL necesitamos todos los parámetros
                if not user or not password or not name:
                    return {
                        "message": "For MySQL/PostgreSQL, 'user', 'password', and 'name' are required"
                    }, 400
                
                app.set_db_config(
                    dbtype=dbtype,
                    user=user,
                    password=password,
                    host=host,
                    port=port,
                    name=name
                )
            else:
                return {
                    "message": f"Invalid database type: {dbtype}. Supported types: sqlite, mysql, postgresql"
                }, 400

            # Conectar a la base de datos
            success = app.connect_to_db(reload=reload)
            
            if success:
                return {
                    "message": "Database connected successfully",
                    "connected": True
                }, 200
            else:
                return {
                    "message": "Failed to connect to database",
                    "connected": False
                }, 400

        except Exception as e:
            return {
                "message": f"Error connecting to database: {str(e)}",
                "connected": False
            }, 400

@ns.route('/disconnect')
class DatabaseDisconnectResource(Resource):

    @api.doc(security='apikey', description="Disconnects from the database.")
    @api.response(200, "Disconnection successful")
    @Api.token_required(auth=True)
    def post(self):
        r"""
        Disconnect from database.

        Closes the database connection and stops history logging.
        """
        try:
            app.disconnect_to_db()
            return {
                "message": "Database disconnected successfully",
                "connected": False
            }, 200
        except Exception as e:
            return {
                "message": f"Error disconnecting from database: {str(e)}"
            }, 400




# setup.py
import setuptools
import platform
import os
import shutil
from pathlib import Path

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    with open("requirements.txt", "r") as fh:
        _requirements = fh.read().splitlines()
except FileNotFoundError:
    _requirements = []

version_ns = {}
with open("version.py") as f:
    exec(f.read(), version_ns)
version = version_ns['__version__']

# Copiar HMI construido a automation/hmi/ antes de empaquetar
hmi_dist_path = Path("hmi/dist")
automation_hmi_path = Path("automation/hmi")

# Limpiar directorio anterior si existe
if automation_hmi_path.exists():
    shutil.rmtree(automation_hmi_path)

# Copiar HMI si existe
if hmi_dist_path.exists() and (hmi_dist_path / "index.html").exists():
    automation_hmi_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.copytree(hmi_dist_path, automation_hmi_path)
    print(f"[INFO] HMI copiado a {automation_hmi_path} para empaquetado")

# Configurar package_data
package_data = {
    'automation': ['pages/assets/*']
}

# Agregar archivos del HMI si fueron copiados
if automation_hmi_path.exists():
    hmi_files = []
    for root, dirs, files in os.walk(automation_hmi_path):
        for file in files:
            # Ruta relativa desde automation/hmi
            rel_path = os.path.relpath(os.path.join(root, file), automation_hmi_path)
            hmi_files.append(f'hmi/{rel_path}')
    if hmi_files:
        package_data['automation'].extend(hmi_files)

system_platform = platform.system()
setuptools.setup(
    name="PyAutomationIO",
    version=version,
    author="KnowAI",
    author_email="dev.know.ai@gmail.com",
    description="A python framework to develop automation industrial processes applications and Artificial Intelligence applications for the industrial field",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GNU AFFERO GENERAL PUBLIC LICENSE",
    url="https://github.com/know-ai/PyAutomation",
    package_data=package_data,
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=_requirements,
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Logging",
        "Topic :: System :: Monitoring"
    ]
)



import pytz
from datetime import datetime, timedelta
from flask_restx import Namespace, Resource, fields, reqparse
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api
from .... import _TIMEZONE, TIMEZONE
from ....variables import VARIABLES

ns = Namespace('Tags', description='Tag Management and Real-time Data')
app = PyAutomation()

query_trends_model = api.model("query_trends_model",{
    'tags':  fields.List(fields.String(), required=True, description='List of tag names to query'),
    'greater_than_timestamp': fields.DateTime(required=True, default=datetime.now(pytz.utc).astimezone(TIMEZONE) - timedelta(minutes=30), description='Start DateTime'),
    'less_than_timestamp': fields.DateTime(required=True, default=datetime.now(pytz.utc).astimezone(TIMEZONE), description='End DateTime'),
    'timezone': fields.String(required=True, default=_TIMEZONE, description='Timezone for the query')
})

query_table_model = api.model("query_table_model",{
    'tags':  fields.List(fields.String(), required=True, description='List of tag names to query'),
    'greater_than_timestamp': fields.DateTime(required=True, default=datetime.now(pytz.utc).astimezone(TIMEZONE) - timedelta(minutes=30), description='Start DateTime'),
    'less_than_timestamp': fields.DateTime(required=True, default=datetime.now(pytz.utc).astimezone(TIMEZONE), description='End DateTime'),
    'timezone': fields.String(required=True, default=_TIMEZONE, description='Timezone for the query'),
    'page': fields.Integer(required=False, default=1, description='Page number'),
    'limit': fields.Integer(required=False, default=20, description='Items per page')
})

query_tabular_data_model = api.model("query_tabular_data_model",{
    'tags':  fields.List(fields.String(), required=True, description='List of tag names to query'),
    'greater_than_timestamp': fields.DateTime(required=True, default=datetime.now(pytz.utc).astimezone(TIMEZONE) - timedelta(minutes=30), description='Start DateTime'),
    'less_than_timestamp': fields.DateTime(required=True, default=datetime.now(pytz.utc).astimezone(TIMEZONE), description='End DateTime'),
    'sample_time': fields.Integer(required=False, default=30, description='Resampling interval in seconds (must be > 0)'),
    'timezone': fields.String(required=True, default=_TIMEZONE, description='Timezone for the query'),
    'page': fields.Integer(required=False, default=1, description='Page number'),
    'limit': fields.Integer(required=False, default=20, description='Items per page')
})

write_value_model = api.model("write_value_model", {
    'tag_name': fields.String(required=True, description='Tag Name'),
    'value': fields.Raw(required=True, description='Value to write (float, int, bool, str)')
})

create_tag_model = api.model("create_tag_model", {
    'name': fields.String(required=True, description='Unique tag name'),
    'unit': fields.String(required=True, description='Engineering unit'),
    'variable': fields.String(required=True, description='Variable type (e.g., Pressure, Temperature)'),
    'display_unit': fields.String(required=False, description='Unit for display purposes', default=''),
    'data_type': fields.String(required=False, description='Data type (float, int, bool, str)', default='float'),
    'description': fields.String(required=False, description='Tag description'),
    'display_name': fields.String(required=False, description='Friendly name for display'),
    'opcua_address': fields.String(required=False, description='OPC UA server URL'),
    'node_namespace': fields.String(required=False, description='OPC UA Node ID'),
    'scan_time': fields.Integer(required=False, description='Polling interval in ms'),
    'dead_band': fields.Float(required=False, description='Deadband value'),
    'process_filter': fields.Boolean(required=False, description='Enable process filter', default=False),
    'gaussian_filter': fields.Boolean(required=False, description='Enable Gaussian filter', default=False),
    'gaussian_filter_threshold': fields.Float(required=False, description='Gaussian filter threshold', default=1.0),
    'gaussian_filter_r_value': fields.Float(required=False, description='Gaussian filter R value', default=0.0),
    'outlier_detection': fields.Boolean(required=False, description='Enable outlier detection', default=False),
    'out_of_range_detection': fields.Boolean(required=False, description='Enable out of range detection', default=False),
    'frozen_data_detection': fields.Boolean(required=False, description='Enable frozen data detection', default=False),
    'segment': fields.String(required=False, description='Network segment', default=''),
    'manufacturer': fields.String(required=False, description='Device manufacturer', default=''),
    'kp': fields.Float(required=False, description='Kilometer Post (KP) value')
})

update_tag_model = api.model("update_tag_model", {
    'id': fields.String(required=True, description='Tag ID'),
    'name': fields.String(required=False, description='Tag name'),
    'unit': fields.String(required=False, description='Engineering unit'),
    'variable': fields.String(required=False, description='Variable type'),
    'display_unit': fields.String(required=False, description='Unit for display purposes'),
    'data_type': fields.String(required=False, description='Data type'),
    'description': fields.String(required=False, description='Tag description'),
    'display_name': fields.String(required=False, description='Friendly name for display'),
    'opcua_address': fields.String(required=False, description='OPC UA server URL'),
    'node_namespace': fields.String(required=False, description='OPC UA Node ID'),
    'scan_time': fields.Integer(required=False, description='Polling interval in ms'),
    'dead_band': fields.Float(required=False, description='Deadband value'),
    'process_filter': fields.Boolean(required=False, description='Enable process filter'),
    'gaussian_filter': fields.Boolean(required=False, description='Enable Gaussian filter'),
    'gaussian_filter_threshold': fields.Float(required=False, description='Gaussian filter threshold'),
    'gaussian_filter_r_value': fields.Float(required=False, description='Gaussian filter R value'),
    'outlier_detection': fields.Boolean(required=False, description='Enable outlier detection'),
    'out_of_range_detection': fields.Boolean(required=False, description='Enable out of range detection'),
    'frozen_data_detection': fields.Boolean(required=False, description='Enable frozen data detection'),
    'segment': fields.String(required=False, description='Network segment'),
    'manufacturer': fields.String(required=False, description='Device manufacturer'),
    'kp': fields.Float(required=False, description='Kilometer Post (KP) value')
})

tags_list_filter_model = api.model("tags_list_filter_model", {
    'manufacturer': fields.String(required=False, description='Filter by manufacturer'),
    'segment': fields.String(required=False, description='Filter by segment')
})

tags_kp_range_model = api.model("tags_kp_range_model", {
    'kp_min': fields.Float(required=True, description='Lower bound KP (inclusive)'),
    'kp_max': fields.Float(required=True, description='Upper bound KP (inclusive)')
})


@ns.route('/')
class TagsCollection(Resource):

    parser = reqparse.RequestParser()
    parser.add_argument('page', type=int, location='args', help='Page number', default=1)
    parser.add_argument('limit', type=int, location='args', help='Items per page', default=20)

    @api.doc(security='apikey', description="Retrieves all available tags with pagination support.")
    @api.response(200, "Success")
    @ns.expect(parser)
    @Api.token_required(auth=True)
    def get(self):
        """
        Get all tags.

        Retrieves a paginated list of all tags currently defined in the system.
        Supports pagination via query parameters: page (default: 1) and limit (default: 20).
        """
        args = self.parser.parse_args()
        page = args.get('page', 1)
        limit = args.get('limit', 20)
        
        # Validate pagination parameters
        if page < 1:
            return {'message': 'Page number must be greater than 0'}, 400
        if limit < 1:
            return {'message': 'Limit must be greater than 0'}, 400
        
        # Get all tags
        all_tags = app.get_tags()
        total = len(all_tags)
        
        # Calculate pagination
        start_idx = (page - 1) * limit
        end_idx = start_idx + limit
        paginated_tags = all_tags[start_idx:end_idx]
        
        return {
            'data': paginated_tags,
            'pagination': {
                'page': page,
                'limit': limit,
                'total': total,
                'pages': (total + limit - 1) // limit if total > 0 else 0
            }
        }, 200

@ns.route('/list')
class TagsListCollection(Resource):

    @api.doc(security='apikey', description="Retrieves a non-paginated list of tags with name, display_unit, variable, display_name. Optional filter by manufacturer and/or segment.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    @ns.expect(tags_list_filter_model)
    def post(self):
        """
        Get tags list (slim).

        Returns a list of tags with only name, display_unit, variable, display_name.
        Accepts optional body with manufacturer and/or segment to filter.
        """
        payload = api.payload or {}
        manufacturer = payload.get('manufacturer')
        segment = payload.get('segment')
        data = app.get_tags_list(manufacturer=manufacturer, segment=segment)
        return {'data': data}, 200

@ns.route('/kp_range')
class TagsByKpRangeCollection(Resource):

    @api.doc(security='apikey', description="Retrieves tags whose KP is between two bounds (inclusive).")
    @api.response(200, "Success")
    @api.response(400, "Invalid payload")
    @Api.token_required(auth=True)
    @ns.expect(tags_kp_range_model)
    def post(self):
        """
        Get tags by KP range.

        Returns tags that have `kp` between `kp_min` and `kp_max` (inclusive).
        """
        payload = api.payload or {}
        kp_min = payload.get('kp_min')
        kp_max = payload.get('kp_max')

        if kp_min is None or kp_max is None:
            return {'message': 'kp_min and kp_max are required'}, 400

        data = app.get_tags_by_kp_range(kp_min=kp_min, kp_max=kp_max)
        return {'data': data}, 200

@ns.route('/names')
class TagsNamesCollection(Resource):

    parser = reqparse.RequestParser()
    parser.add_argument('names', type=str, action='append', location='args', help='List of tag names to retrieve')

    @api.doc(security='apikey', description="Retrieves specific tags by name.")
    @api.response(200, "Success")
    @ns.expect(parser)
    @Api.token_required(auth=True)
    def get(self):
        """
        Get tags by name.

        Retrieves details for a specific list of tags provided as query parameters.
        """
        args = self.parser.parse_args()
        names = args.get('names')
        return app.get_tags_by_names(names=names or []), 200
    
@ns.route('/query_trends')
class QueryTrendsResource(Resource):

    @api.doc(security='apikey', description="Queries historical trend data for tags.")
    @api.response(200, "Success")
    @api.response(400, "Invalid parameters or Timezone")
    @api.response(404, "Tag not found")
    @Api.token_required(auth=True)
    @ns.expect(query_trends_model)
    def post(self):
        """
        Query trends.

        Retrieves historical time-series data for a list of tags within a specified time range.
        
        Authorized Roles: {0}
        """
        timezone = _TIMEZONE
        tags = api.payload['tags']
        if "timezone" in api.payload:

            timezone = api.payload["timezone"]

        if timezone not in pytz.all_timezones:

            return f"Invalid Timezone", 400
        
        for tag in tags:

            if not app.get_tag_by_name(name=tag):

                return f"{tag} not exist into db", 404
        
        separator = '.'
        greater_than_timestamp = api.payload['greater_than_timestamp']
        start = greater_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        less_than_timestamp = api.payload['less_than_timestamp']
        stop = less_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        result = app.get_trends(start, stop, timezone, *tags)
        
        return result, 200

@ns.route('/query_table')
class QueryTableResource(Resource):

    @api.doc(security='apikey', description="Queries historical data in table format.")
    @api.response(200, "Success")
    @api.response(400, "Invalid parameters")
    @api.response(404, "Tag not found")
    @Api.token_required(auth=True)
    @ns.expect(query_table_model)
    def post(self):
        """
        Query data table.

        Retrieves historical tag values in a paginated list format.

        Authorized Roles: {0}
        """
        timezone = _TIMEZONE
        tags = api.payload['tags']
        page = api.payload.get('page', 1)
        limit = api.payload.get('limit', 20)

        if "timezone" in api.payload:
            timezone = api.payload["timezone"]

        if timezone not in pytz.all_timezones:
            return f"Invalid Timezone", 400
        
        for tag in tags:
            if not app.get_tag_by_name(name=tag):
                return f"{tag} not exist into db", 404
        
        separator = '.'
        greater_than_timestamp = api.payload['greater_than_timestamp']
        # Ensure timestamp format is consistent
        start = greater_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        less_than_timestamp = api.payload['less_than_timestamp']
        stop = less_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        result = app.get_tags_tables(start, stop, timezone, tags, page, limit)
        
        return result, 200
    
@ns.route('/get_tabular_data')
class GetTabularDataResource(Resource):

    @api.doc(security='apikey', description="Queries historical data in a resampled tabular format.")
    @api.response(200, "Success")
    @api.response(400, "Invalid parameters")
    @api.response(404, "Tag not found")
    @Api.token_required(auth=True)
    @ns.expect(query_tabular_data_model)
    def post(self):
        """
        Get tabular data.

        Query tag values in tabular format with pagination and resampling.
        
        The result contains data points at regular intervals (sample_time) from greater_than_timestamp 
        up to less_than_timestamp. If exact data is missing, the previous known value is used (forward fill).
        
        Authorized Roles: {0}
        """
        timezone = _TIMEZONE
        tags = api.payload['tags']
        page = api.payload.get('page', 1)
        limit = api.payload.get('limit', 20)
        sample_time = api.payload.get('sample_time', 30)

        # Validar que sample_time sea un entero positivo > 0
        if not isinstance(sample_time, int) or sample_time <= 0:
            return {'message': 'sample_time must be a positive integer greater than 0'}, 400

        if "timezone" in api.payload:
            timezone = api.payload["timezone"]

        if timezone not in pytz.all_timezones:
            return f"Invalid Timezone", 400
        
        for tag in tags:
            if not app.get_tag_by_name(name=tag):
                return f"{tag} not exist into db", 404
        
        separator = '.'
        greater_than_timestamp = api.payload['greater_than_timestamp']
        # Ensure timestamp format is consistent
        start = greater_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        less_than_timestamp = api.payload['less_than_timestamp']
        stop = less_than_timestamp.replace("T", " ").split(separator, 1)[0] + '.00'
        
        result = app.get_tabular_data(start, stop, timezone, tags, sample_time, page, limit)
        
        return result, 200

@ns.route('/write_value')
class WriteValueResource(Resource):

    @api.doc(security='apikey', description="Writes a value to a tag.")
    @api.response(200, "Success (CVT and OPC UA if applicable)")
    @api.response(207, "Partial Success (CVT OK, OPC UA Failed)")
    @api.response(404, "Tag not found")
    @api.response(500, "Internal Error")
    @Api.token_required(auth=True)
    @ns.expect(write_value_model)
    def post(self):
        """
        Write tag value.

        Writes a value to a tag in the Current Value Table (CVT). 
        If the tag is mapped to an OPC UA Node, it also attempts to write to the OPC UA Server.
        
        Authorized Roles: {0}
        """
        tag_name = api.payload['tag_name']
        value = api.payload['value']
        
        # Buscar el tag en CVT
        tag = app.cvt.get_tag_by_name(name=tag_name)
        if not tag:
            return {'message': f'Tag {tag_name} does not exist', 'success': False}, 404
        
        # Escribir en CVT
        try:
            timestamp = datetime.now(pytz.utc).astimezone(TIMEZONE)
            app.cvt.set_value(id=tag.id, value=value, timestamp=timestamp)
        except Exception as err:
            return {
                'message': f'Error writing to CVT: {str(err)}',
                'tag': tag_name,
                'success': False
            }, 500
        
        # Si tiene node_namespace, escribir en OPC UA Server usando el método de core
        opcua_result = None
        opcua_status = None
        if tag.node_namespace and tag.opcua_address:
            opcua_result, opcua_status = app.write_opcua_value(
                opcua_address=tag.opcua_address,
                node_namespace=tag.node_namespace,
                value=value
            )
        
        # Resultado consolidado
        result = {
            'message': 'Value written to CVT' + (' and OPC UA' if opcua_status == 200 else ''),
            'tag': tag_name,
            'value': value,
            'cvt_success': True,
            'opcua_success': opcua_status == 200 if opcua_status else None,
            'opcua_detail': opcua_result if opcua_result else None
        }
        
        # Status: 200 si CVT OK, aunque OPC UA falle (parcial success)
        final_status = 200 if opcua_status in (200, None) else 207  # 207 = Multi-Status
        return result, final_status

@ns.route('/add')
class AddTagResource(Resource):

    @api.doc(security='apikey', description="Creates a new tag in the system.")
    @api.response(200, "Tag created successfully")
    @api.response(400, "Tag creation failed")
    @Api.token_required(auth=True)
    @ns.expect(create_tag_model)
    def post(self):
        """
        Create tag.

        Creates a new tag in the automation application with the specified configuration.
        """
        payload = api.payload
        
        # Required fields
        name = payload.get('name')
        unit = payload.get('unit')
        variable = payload.get('variable')
        
        if not name or not unit or not variable:
            return {
                'message': 'Missing required fields: name, unit, and variable are required'
            }, 400
        
        try:
            tag, message = app.create_tag(
                name=name,
                unit=unit,
                variable=variable,
                display_unit=payload.get('display_unit', ''),
                data_type=payload.get('data_type', 'float'),
                description=payload.get('description'),
                display_name=payload.get('display_name'),
                opcua_address=payload.get('opcua_address'),
                node_namespace=payload.get('node_namespace'),
                scan_time=payload.get('scan_time'),
                dead_band=payload.get('dead_band'),
                process_filter=payload.get('process_filter', False),
                gaussian_filter=payload.get('gaussian_filter', False),
                gaussian_filter_threshold=payload.get('gaussian_filter_threshold', 1.0),
                gaussian_filter_r_value=payload.get('gaussian_filter_r_value', 0.0),
                outlier_detection=payload.get('outlier_detection', False),
                out_of_range_detection=payload.get('out_of_range_detection', False),
                frozen_data_detection=payload.get('frozen_data_detection', False),
                segment=payload.get('segment', ''),
                manufacturer=payload.get('manufacturer', ''),
                kp=payload.get('kp')
            )
            
            if tag:
                return {
                    'message': f"Tag '{name}' created successfully",
                    'tag': {
                        'id': tag.id,
                        'name': tag.name,
                        'unit': tag.unit,
                        'variable': tag.variable
                    }
                }, 200
            else:
                return {
                    'message': f"Failed to create tag: {message}"
                }, 400
        except Exception as e:
            return {
                'message': f"Error creating tag: {str(e)}"
            }, 400


@ns.route('/update')
class UpdateTagResource(Resource):

    @api.doc(security='apikey', description="Updates an existing tag configuration.")
    @api.response(200, "Tag updated successfully")
    @api.response(400, "Tag update failed")
    @api.response(404, "Tag not found")
    @Api.token_required(auth=True)
    @ns.expect(update_tag_model)
    def post(self):
        """
        Update tag.

        Updates the configuration of an existing tag. Only provided fields will be updated.
        """
        payload = api.payload
        
        # Required field
        tag_id = payload.get('id')
        if not tag_id:
            return {
                'message': 'Tag ID is required'
            }, 400
        
        # Check if tag exists
        try:
            tag = app.cvt.get_tag(id=tag_id)
            if not tag:
                return {
                    'message': f'Tag with ID {tag_id} not found'
                }, 404
        except Exception:
            return {
                'message': f'Tag with ID {tag_id} not found'
            }, 404
        
        # Build kwargs with only provided fields (excluding 'id')
        update_kwargs = {k: v for k, v in payload.items() if k != 'id' and v is not None}
        
        if not update_kwargs:
            return {
                'message': 'No fields to update provided'
            }, 400
        
        try:
            updated_tag, message = app.update_tag(id=tag_id, **update_kwargs)
            
            if updated_tag:
                return {
                    'message': f"Tag '{updated_tag.name}' updated successfully",
                    'tag': {
                        'id': updated_tag.id,
                        'name': updated_tag.name
                    }
                }, 200
            else:
                return {
                    'message': f"Failed to update tag: {message}"
                }, 400
        except Exception as e:
            return {
                'message': f"Error updating tag: {str(e)}"
            }, 400


@ns.route('/delete/<tag_name>')
@api.param('tag_name', 'The tag name to delete')
class DeleteTagResource(Resource):

    @api.doc(security='apikey', description="Deletes a tag from the system by name.")
    @api.response(200, "Tag deleted successfully")
    @api.response(400, "Tag deletion failed")
    @api.response(404, "Tag not found")
    @Api.token_required(auth=True)
    def delete(self, tag_name):
        """
        Delete tag.

        Deletes a tag from the system by its name. 
        Note: Tags with associated alarms cannot be deleted.
        """
        # Check if tag exists
        tag = app.get_tag_by_name(name=tag_name)
        if not tag:
            return {
                'message': f'Tag "{tag_name}" not found'
            }, 404
        
        try:
            result = app.delete_tag_by_name(name=tag_name)
            
            if result:
                # If result is a string, it's an error message
                return {
                    'message': result
                }, 400
            else:
                return {
                    'message': f'Tag "{tag_name}" deleted successfully'
                }, 200
        except Exception as e:
            return {
                'message': f"Error deleting tag: {str(e)}"
            }, 400


@ns.route('/timezones')
class TimezonesCollection(Resource):

    @api.doc(security='apikey', description="Retrieves a list of available timezones.")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self):
        """
        Get Timezones.

        Retrieves a list of all supported timezones.
        """
        return pytz.all_timezones, 200


@ns.route('/variables')
class VariablesCollection(Resource):

    @api.doc(security='apikey', description="Retrieves a list of all available variables (e.g., Pressure, Temperature).")
    @api.response(200, "Success")
    @Api.token_required(auth=True)
    def get(self):
        """
        Get Variables.

        Retrieves a list of all available physical variables in the system.
        Examples: Pressure, Temperature, Length, Mass, etc.
        """
        # VARIABLES is a dict where keys are variable names
        variables = list(VARIABLES.keys())
        return {
            'data': variables,
            'total': len(variables)
        }, 200


@ns.route('/units/<variable_name>')
@api.param('variable_name', 'The variable name (e.g., Pressure, Temperature)')
class UnitsByVariableResource(Resource):

    @api.doc(security='apikey', description="Retrieves available units for a specific variable.")
    @api.response(200, "Success")
    @api.response(404, "Variable not found")
    @Api.token_required(auth=True)
    def get(self, variable_name):
        """
        Get Units by Variable.

        Retrieves all available units for a specific variable.
        The variable_name must match one of the available variables (case-sensitive).
        
        Example: GET /tags/units/Pressure returns units like ['bar', 'mbar', 'Pa', 'kPa', etc.]
        """
        # Check if variable exists
        if variable_name not in VARIABLES:
            return {
                'message': f'Variable "{variable_name}" not found. Available variables: {list(VARIABLES.keys())}'
            }, 404
        
        # Get units for this variable
        # VARIABLES[variable_name] is a dict where values are unit symbols
        units_dict = VARIABLES[variable_name]
        units = list(units_dict.values())
        
        return {
            'variable': variable_name,
            'data': units,
            'total': len(units)
        }, 200


from flask import request
from flask_restx import Namespace, Resource, fields
from .... import PyAutomation
from ....extensions.api import api
from ....extensions import _api as Api


ns = Namespace('OPCUA Server', description='OPC UA Server Management Resources')
app = PyAutomation()

# Models
update_access_type_model = api.model("update_access_type_model", {
    'namespace': fields.String(required=True, description='OPC UA node namespace string'),
    'access_type': fields.String(required=True, description='Access type: Read, Write, or ReadWrite'),
    'name': fields.String(required=False, description='Node name (optional, used if record does not exist)')
})


@ns.route('/attrs')
class OPCUAServerAttributesResource(Resource):

    @api.doc(security='apikey', description="Retrieves all attributes (variables and properties) from the OPC UA Server state machine.")
    @api.response(200, "Success")
    @api.response(404, "OPC UA Server not found")
    @Api.token_required(auth=True)
    def get(self):
        r"""
        Get OPC UA Server attributes.

        Retrieves all OPC UA nodes (variables and their properties) from the embedded OPC UA Server
        with their access levels (Read, Write, ReadWrite).

        Returns a list of dictionaries containing:
        - name: Full path name (parent_folder.variable_name or parent_folder.variable_name.property_name)
        - namespace: OPC UA node namespace string
        - access_type: Access level ("Read", "Write", or "ReadWrite")
        """
        try:
            attrs = app.get_opcua_server_attrs()
            return {
                "data": attrs
            }, 200
        except Exception as e:
            return {
                "message": f"Failed to retrieve OPC UA Server attributes: {str(e)}"
            }, 404


@ns.route('/attrs/update')
class OPCUAServerUpdateAccessTypeResource(Resource):

    @api.doc(security='apikey', description="Updates the access type (Read, Write, ReadWrite) for a specific OPC UA Server node.")
    @api.response(200, "Access type updated successfully")
    @api.response(400, "Invalid request or parameters")
    @api.response(404, "Node not found")
    @Api.token_required(auth=True)
    @ns.expect(update_access_type_model)
    def put(self):
        r"""
        Update OPC UA Server node access type.

        Updates the access level for a specific OPC UA Server node identified by its namespace.
        The access_type must be one of: "Read", "Write", or "ReadWrite".

        Request body:
        - namespace: OPC UA node namespace string (required)
        - access_type: New access type - "Read", "Write", or "ReadWrite" (required)
        - name: Node name (optional, used if database record does not exist)
        """
        if not request.is_json:
            return {
                "message": "Request must be JSON"
            }, 400
        
        data = request.json
        namespace = data.get('namespace')
        access_type = data.get('access_type')
        name = data.get('name')
        
        if not namespace:
            return {
                "message": "namespace parameter is required"
            }, 400
        
        if not access_type:
            return {
                "message": "access_type parameter is required"
            }, 400
        
        if access_type not in ["Read", "Write", "ReadWrite"]:
            return {
                "message": "access_type must be one of: 'Read', 'Write', or 'ReadWrite'"
            }, 400
        
        try:
            success, message = app.update_opcua_server_node_access_type(
                namespace=namespace,
                access_type=access_type,
                name=name
            )
            
            if success:
                return {
                    "message": message,
                    "namespace": namespace,
                    "access_type": access_type
                }, 200
            else:
                return {
                    "message": message
                }, 404
        except Exception as e:
            return {
                "message": f"Failed to update access type: {str(e)}"
            }, 400




import json
import unittest
from datetime import datetime


class FakeRange:
    def __init__(self, low, high):
        self.Low = low
        self.High = high


class TestOpcuaSerialization(unittest.TestCase):
    def test_to_jsonable_range(self):
        # Import local para no romper discovery si dependencias opcua no están cargadas al importar tests
        from ..opcua.models import Client

        r = FakeRange(low=0.0, high=100.0)
        out = Client._to_jsonable(r)
        self.assertIsInstance(out, dict)
        self.assertEqual(out["Low"], 0.0)
        self.assertEqual(out["High"], 100.0)
        # Debe ser serializable por json
        json.dumps(out)

    def test_to_jsonable_datetime(self):
        from ..opcua.models import Client

        ts = datetime(2026, 1, 1, 12, 0, 0)
        out = Client._to_jsonable(ts)
        self.assertEqual(out, ts.isoformat())
        json.dumps(out)

    def test_to_jsonable_nested(self):
        from ..opcua.models import Client

        payload = {
            "a": FakeRange(1, 2),
            "b": [FakeRange(3, 4), None],
        }
        out = Client._to_jsonable(payload)
        json.dumps(out)





from .users import Role, Roles, Users

# Init Resources
def init_app(app):

    from ..modules.tags.resources import init_app as init_tags
    from ..modules.alarms.resources import init_app as init_alarms
    from ..modules.users.resources import init_app as init_users
    from ..modules.events.resources import init_app as init_events
    from ..modules.settings.resources import init_app as init_settings
    from ..modules.opcua.resources import init_app as init_opcua
    from ..modules.database.resources import init_app as init_database
    from ..modules.machines.resources import init_app as init_machines
    from ..modules.health.resources import init_app as init_health
    from ..modules.linear_referencing.resources import init_app as init_linear_referencing

    init_tags()
    init_alarms()
    init_users()
    init_events()
    init_settings()
    init_opcua()
    init_database()
    init_machines()
    init_health()
    init_linear_referencing()